/* A base class for all repositories of directories of source files.

Metropolis: Design Environment for Heterogeneus Systems.

Copyright (c) 1998-2005 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in all
copies of this software and that appropriate acknowledgments are made
to the research of the Metropolis group.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.


METROPOLIS_COPYRIGHT_VERSION_1
COPYRIGHTENDKEY


*/


package metropolis.metamodel.frontend;

import java.io.File;

//////////////////////////////////////////////////////////////////////////
//// DirRepository
/**
 * A base class for all repositories that store information about a
 * directory containing source files. It provides utility methods to check
 * if a repository is still valid, and methods to get the name of the
 * repository.
 * <p>
 * Portions of this code were derived from sources developed under the
 * auspices of the Ptolemy II project.
 *
 * @author   Robert Clariso
 * @version  $Id: DirRepository.java,v 1.13.4.1 2005/02/14 18:46:38 cxh Exp $
 */
public abstract class DirRepository extends Repository {

    ///////////////////////////////////////////////////////////////////
    ////                           constructors                    ////

    /** Create a new repository with information on the sources in a
     *  given source directory. The source path must be a directory.
     *  @param src Path where the source file is stored.
     */
    public DirRepository(File src) {
        super(src);
        if (!src.isDirectory()) {
            throw new RuntimeException(src + " is not a directory");
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                          public methods                   ////

    /** Check if it is possible to update the repository instead of
     *  rebuilding it from scratch. In the case of repositories based
     *  on directories, in most case we will be able to reuse information
     *  from the files that have not been modified.
     *  @return true iff the repository can be updated.
     */
    public boolean canUpdate() { return true; }


    ///////////////////////////////////////////////////////////////////
    ////                         protected methods                 ////

    /** Compute the path where a repository of a set of sources will be
     *  stored. The path of the repository depends on the path where
     *  the sources e.g. file(s), are stored.
     *  @param src Path where the sources are stored.
     *  @return The path where the repository file will be stored.
     */
    protected File _repositoryPath(File src) {
        // The name of the repository will be:
        //  (source)      path /
        //  (repository)  path / _name
        return new File (src, _getRepositoryName());
    }

    /** Check if this repository exists and is completely up-to-date.
     *  @return true iff the repository exists and the sources have
     *  not changed since the repository was created.
     */
    protected boolean isValid() {
        if (!_rep.exists()) return false;
        File[] files = _src.listFiles();
        for (int i = 0; i < files.length; i++) {
            File file = files[i];
            if (isSource(file) &&
                    (file.lastModified() > _rep.lastModified())) {
                System.out.println("File " + file + " is newer");
                return false;
            }
        }
        return true;
    }

    /** Return true if a file is a source file, i.e. it is a directory
     *  or it is file that ends with the meta-model extension.
     *  @param file Possible source file.
     *  @return true if the file is a source file.
     */
    protected boolean isSource(File file) {
        if (file.isDirectory()) return false;
        String name = file.getName();
        return name.endsWith("." + FileLoader._metaModelExtension);
    }

    /** Return the name of the repository file for this directory.
     *  @return The name of this repository file.
     */
    protected abstract String _getRepositoryName();

}




